from __future__ import unicode_literals

import logging

from lunr.__main__ import lunr

__all__ = ("lunr",)

logging.basicConfig(format="%(levelname)-7s -  %(message)s")

__VERSION__ = "0.5.8"
__TARGET_JS_VERSION__ = "2.3.8"



from __future__ import unicode_literals

from builtins import str

import pytest

from lunr.field_ref import FieldRef
from lunr.exceptions import BaseLunrException


class TestFieldRef:
    def test_str_combines_document_ref_and_field_name(self):
        field_name = "title"
        document_ref = 123
        field_ref = FieldRef(document_ref, field_name)

        assert str(field_ref) == "title/123"
        assert repr(field_ref) == '<FieldRef field="title" ref="123">'

    def test_from_string_splits_string_into_parts(self):
        field_ref = FieldRef.from_string("title/123")

        assert field_ref.field_name == "title"
        assert field_ref.doc_ref == "123"

    def test_from_string_docref_contains_join_character(self):
        field_ref = FieldRef.from_string("title/http://example.com/123")

        assert field_ref.field_name == "title"
        assert field_ref.doc_ref == "http://example.com/123"

    def test_from_string_does_not_contain_join_character(self):
        string = "docRefOnly"

        with pytest.raises(BaseLunrException):
            FieldRef.from_string(string)



from __future__ import unicode_literals

import six

from lunr.query_lexer import QueryLexer
from lunr.query import Clause, QueryPresence
from lunr.exceptions import QueryParseError


class QueryParser:
    def __init__(self, string, query):
        self.lexer = QueryLexer(string)
        self.query = query
        self.current_clause = Clause()
        self.lexeme_idx = 0

    def parse(self):
        self.lexer.run()
        self.lexemes = self.lexer.lexemes

        state = self.__class__.parse_clause

        while state:
            state = state(self)

        return self.query

    def peek_lexeme(self):
        try:
            return self.lexemes[self.lexeme_idx]
        except IndexError:
            return None

    def consume_lexeme(self):
        lexeme = self.peek_lexeme()
        self.lexeme_idx += 1
        return lexeme

    def next_clause(self):
        self.query.clause(self.current_clause)
        self.current_clause = Clause()

    @classmethod
    def parse_clause(cls, parser):
        lexeme = parser.peek_lexeme()
        if lexeme is None:
            return

        if lexeme["type"] == QueryLexer.FIELD:
            return cls.parse_field
        elif lexeme["type"] == QueryLexer.TERM:
            return cls.parse_term
        elif lexeme["type"] == QueryLexer.PRESENCE:
            return cls.parse_presence
        else:
            raise QueryParseError(
                "Expected either a field or a term, found {}{}".format(
                    lexeme["type"],
                    'with value "' + lexeme["string"] + '"'
                    if len(lexeme["string"])
                    else "",
                )
            )

    @classmethod
    def parse_field(cls, parser):
        lexeme = parser.consume_lexeme()

        if lexeme["string"] not in parser.query.all_fields:
            raise QueryParseError(
                'Unrecognized field "{}", possible fields {}'.format(
                    lexeme["string"], ", ".join(parser.query.all_fields)
                )
            )

        parser.current_clause.fields = [lexeme["string"]]

        next_lexeme = parser.peek_lexeme()
        if next_lexeme is None:
            raise QueryParseError("Expected term, found nothing")

        if next_lexeme["type"] == QueryLexer.TERM:
            return cls.parse_term
        else:
            raise QueryParseError("Expected term, found {}".format(next_lexeme["type"]))

    @classmethod
    def parse_term(cls, parser):
        lexeme = parser.consume_lexeme()

        parser.current_clause.term = lexeme["string"].lower()
        if "*" in lexeme["string"]:
            parser.current_clause.use_pipeline = False

        return cls._peek_next_lexeme(parser)

    @classmethod
    def parse_presence(cls, parser):
        lexeme = parser.consume_lexeme()

        if lexeme is None:
            return

        if lexeme["string"] == "-":
            parser.current_clause.presence = QueryPresence.PROHIBITED
        elif lexeme["string"] == "+":
            parser.current_clause.presence = QueryPresence.REQUIRED
        else:
            raise QueryParseError(
                "Unrecognized parser operator: {}, expected `+` or `-`".format(
                    lexeme.str
                )
            )

        next_lexeme = parser.peek_lexeme()
        if next_lexeme is None:
            raise QueryParseError("Expected either a field or a term, found nothing")

        if next_lexeme["type"] == QueryLexer.FIELD:
            return cls.parse_field
        elif next_lexeme["type"] == QueryLexer.TERM:
            return cls.parse_term
        else:
            raise QueryParseError(
                "Expected either a field or a term, found {}".format(lexeme["type"])
            )

    @classmethod
    def parse_edit_distance(cls, parser):
        lexeme = parser.consume_lexeme()

        try:
            edit_distance = int(lexeme["string"])
        except ValueError as e:
            six.raise_from(QueryParseError("Edit distance must be numeric"), e)

        parser.current_clause.edit_distance = edit_distance

        return cls._peek_next_lexeme(parser)

    @classmethod
    def parse_boost(cls, parser):
        lexeme = parser.consume_lexeme()

        try:
            boost = int(lexeme["string"])
        except ValueError as e:
            six.raise_from(QueryParseError("Boost must be numeric"), e)

        parser.current_clause.boost = boost

        return cls._peek_next_lexeme(parser)

    @classmethod
    def _peek_next_lexeme(cls, parser):
        next_lexeme = parser.peek_lexeme()
        if next_lexeme is None:
            parser.next_clause()
            return

        if next_lexeme["type"] == QueryLexer.TERM:
            parser.next_clause()
            return cls.parse_term
        elif next_lexeme["type"] == QueryLexer.FIELD:
            parser.next_clause()
            return cls.parse_field
        elif next_lexeme["type"] == QueryLexer.EDIT_DISTANCE:
            return cls.parse_edit_distance
        elif next_lexeme["type"] == QueryLexer.BOOST:
            return cls.parse_boost
        elif next_lexeme["type"] == QueryLexer.PRESENCE:
            parser.next_clause()
            return cls.parse_presence
        else:
            raise QueryParseError(
                "Unexpected lexeme type {}".format(next_lexeme["type"])
            )



from __future__ import unicode_literals


from enum import Enum


class QueryPresence(Enum):
    """Defines possible behaviours for the term's presence in a document."""

    OPTIONAL = 1  # default
    REQUIRED = 2
    PROHIBITED = 3  # documents that contain this term will not be returned


class Query(object):
    """A `lunr.Query` provides a programmatic way of defining queries to be
    performed against a `lunr.Index`.

    Prefer constructing a `lunr.Query` using `the lunr.Index.query` method
    so the query object is pre-initialized with the right index fields.
    """

    # Constants for indicating what kind of automatic wildcard insertion will
    # be used when constructing a query clause.
    # This allows wildcards to be added to the beginning and end of a term
    # without having to manually do any string concatenation.
    # The wildcard constants can be bitwise combined to select both leading and
    # trailing wildcards.
    WILDCARD = "*"
    WILDCARD_NONE = 0
    WILDCARD_LEADING = 1
    WILDCARD_TRAILING = 2

    def __init__(self, all_fields):
        self.clauses = []
        self.all_fields = all_fields

    def __repr__(self):
        return '<Query fields="{}" clauses="{}">'.format(
            ",".join(self.all_fields), ",".join(c.term for c in self.clauses)
        )

    def clause(self, *args, **kwargs):
        """Adds a `lunr.Clause` to this query.

        Unless the clause contains the fields to be matched all fields will be
        matched. In addition a default boost of 1 is applied to the clause.

        If the first argument is a `lunr.Clause` it will be mutated and added,
        otherwise args and kwargs will be used in the constructor.

        Returns:
            lunr.Query: The Query itself.
        """
        if args and isinstance(args[0], Clause):
            clause = args[0]
        else:
            clause = Clause(*args, **kwargs)

        if not clause.fields:
            clause.fields = self.all_fields

        if (clause.wildcard & Query.WILDCARD_LEADING) and (
            clause.term[0] != Query.WILDCARD
        ):
            clause.term = Query.WILDCARD + clause.term

        if (clause.wildcard & Query.WILDCARD_TRAILING) and (
            clause.term[-1] != Query.WILDCARD
        ):
            clause.term = clause.term + Query.WILDCARD

        self.clauses.append(clause)
        return self

    def term(self, term, **kwargs):
        """Adds a term to the current query, creating a Clause and adds it to
        the list of clauses making up this Query.

        The term is not tokenized and used "as is". Any conversion to token
        or token-like strings should be performed before calling this method.

        For example:
            query.term(lunr.Tokenizer("foo bar"))

        Args:
            term (Token or iterable): Token or iterable of tokens to add.
            kwargs (dict): Additional properties to add to the Clause.
        """
        if isinstance(term, (list, tuple)):
            for t in term:
                self.term(t, **kwargs)
        else:
            self.clause(str(term), **kwargs)

        return self

    def is_negated(self):
        """A negated query is one in which every clause has a presence of
        prohibited. These queries require some special processing to return
        the expected results.
        """
        return all(
            clause.presence == QueryPresence.PROHIBITED for clause in self.clauses
        )


class Clause(object):
    """A single clause in a `lunr.Query` contains a term and details on
    how to match that term against a `lunr.Index`

    Args:
        term (str, optional): The term for the clause.
        field (iterable, optional): The fields for the term to be searched
            against.
        edit_distance (int, optional): The character distance to use, defaults
            to 0.
        use_pipeline (bool, optional): Whether the clause should be pre
            processed by the index's pipeline, default to True.
        boost (int, optional): Boost to apply to the clause, defaults to 1.
        wildcard (Query.WILDCARD_*, optional): Any of the Query.WILDCARD
            constants defining if a wildcard is to be used and how, defaults
            to Query.WILDCARD_NONE.
        presence (QueryPresence, optional): Behaviour for a terms presence
            in a document.
    """

    def __init__(
        self,
        term=None,
        fields=None,
        edit_distance=0,
        use_pipeline=True,
        boost=1,
        wildcard=Query.WILDCARD_NONE,
        presence=QueryPresence.OPTIONAL,
    ):
        super(Clause, self).__init__()
        self.term = term
        self.fields = fields or []
        self.edit_distance = edit_distance
        self.use_pipeline = use_pipeline
        self.boost = boost
        self.wildcard = wildcard
        self.presence = presence

    def __repr__(self):
        return '<Clause term="{}">'.format(self.term)



from __future__ import unicode_literals

from builtins import str
from copy import deepcopy

from lunr.token import Token
from lunr.utils import as_string

SEPARATOR_CHARS = " \t\n\r\f\v\xa0-"


def default_separator(char):
    return char and char in SEPARATOR_CHARS


def Tokenizer(obj, metadata=None, separator=None):
    """Splits a string into tokens ready to be inserted into the search index.

    Args:
        metadata (dict): Optional metadata can be passed to the tokenizer, this
            metadata will be cloned and added as metadata to every token that is
            created from the object to be tokenized.
        separator (callable or compiled regex): This tokenizer will convert its
            parameter to a string by calling `str` and then will split this
            string on characters for which `separator` is True. Lists will have
            their elements converted to strings and wrapped in a lunr `Token`.

    Returns:
        List of Token instances.
    """
    if obj is None:
        return []

    metadata = metadata or {}

    if isinstance(obj, (list, tuple)):
        return [
            Token(as_string(element).lower(), deepcopy(metadata)) for element in obj
        ]

    if separator is None:
        is_separator = default_separator
    elif callable(separator):
        is_separator = separator
    else:  # must be a regex, remove when dropping support for 2.7
        is_separator = lambda c: separator.match(c)  # noqa

    string = str(obj).lower()
    length = len(string)
    tokens = []
    slice_start = 0
    for slice_end in range(length + 1):
        char = string[slice_end] if slice_end != length else ""
        slice_length = slice_end - slice_start
        if is_separator(char) or slice_end == length:
            if slice_length > 0:
                token_metadata = {}
                token_metadata["position"] = [slice_start, slice_length]
                token_metadata["index"] = len(tokens)
                token_metadata.update(metadata)

                sl = slice(slice_start, slice_end)
                tokens.append(Token(string[sl], token_metadata))

            slice_start = slice_end + 1

    return tokens



import pytest

from lunr import lunr
from lunr.query import Query, QueryPresence
from lunr.exceptions import QueryParseError


class TestSingleTermSearch:
    def test_one_match(self, index):
        results = index.search("scarlett")
        assert len(results) == 1
        assert results[0]["ref"] == "c"
        assert set(results[0]["match_data"].metadata.keys()) == {"scarlett"}

    def test_no_match(self, index):
        results = index.search("foo")
        assert len(results) == 0

    def test_multiple_matches_sorts_by_relevance(self, index):
        results = index.search("plant")
        assert len(results) == 2
        assert results[0]["ref"] == "b"
        assert results[1]["ref"] == "c"

    def test_pipeline_processing_enabled(self, index):
        query = Query(index.fields)
        query.clause(term="study", use_pipeline=True)  # stemmed to studi
        results = index.query(query)

        assert len(results) == 2
        assert results[0]["ref"] == "b"
        assert results[1]["ref"] == "a"

    def test_pipeline_processing_disabled(self, index):
        query = Query(index.fields)
        query.clause(term="study", use_pipeline=False)  # stemmed to studi
        results = index.query(query)

        assert len(results) == 0

    def test_multiple_terms_all_terms_match(self, index):
        results = index.search("fellow candlestick")

        assert len(results) == 1
        assert results[0]["ref"] == "a"
        metadata = results[0]["match_data"].metadata
        assert set(metadata.keys()) == {"fellow", "candlestick"}
        assert set(metadata["fellow"].keys()) == {"body"}
        assert set(metadata["candlestick"].keys()) == {"body"}

    def test_one_term_matches(self, index):
        results = index.search("week foo")

        assert len(results) == 1
        assert results[0]["ref"] == "c"
        metadata = results[0]["match_data"].metadata
        assert set(metadata.keys()) == {"week"}

    def test_duplicate_query_terms(self, index):
        index.search("fellow candlestick foo bar green plant fellow")

    def test_documents_with_all_terms_score_higher(self, index):
        results = index.search("candlestick green")

        assert len(results) == 3

        assert {r["ref"] for r in results} == {"a", "b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"candlestick", "green"}
        assert set(results[1]["match_data"].metadata.keys()) == {"green"}
        assert set(results[2]["match_data"].metadata.keys()) == {"green"}

    def test_no_terms_match(self, index):
        results = index.search("foo bar")

        assert len(results) == 0

    def test_corpus_terms_are_stemmed(self, index):
        results = index.search("water")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}

    def test_field_scoped_terms(self, index):
        results = index.search("title:plant")

        assert len(results) == 1
        assert results[0]["ref"] == "b"
        assert set(results[0]["match_data"].metadata.keys()) == {"plant"}

    def test_field_scoped_no_matching_terms(self, index):
        results = index.search("title:candlestick")

        assert len(results) == 0


class TestSearchWildcardTrailing:
    def test_matching_no_matches(self, index):
        results = index.search("fo*")

        assert len(results) == 0

    def test_matching_one_match(self, index):
        results = index.search("candle*")

        assert len(results) == 1
        assert results[0]["ref"] == "a"
        assert set(results[0]["match_data"].metadata.keys()) == {"candlestick"}

    def test_matching_multiple_terms_match(self, index):
        results = index.search("pl*")

        assert len(results) == 2

        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"plumb", "plant"}
        assert set(results[1]["match_data"].metadata.keys()) == {"plumb", "plant"}


class TestSearchWildcardLeading:
    def test_matching_no_matches(self, index):
        results = index.search("*oo")

        assert len(results) == 0

    def test_matching_multiple_terms_match(self, index):
        results = index.search("*ant")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"plant"}
        assert set(results[1]["match_data"].metadata.keys()) == {"plant"}


class TestSearchWildcardContained:
    def test_matching_no_matches(self, index):
        results = index.search("f*o")
        assert len(results) == 0

    def test_matching_multiple_terms_match(self, index):
        results = index.search("pl*nt")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"plant"}
        assert set(results[1]["match_data"].metadata.keys()) == {"plant"}


class TestEditDistance:
    def test_edit_distance_no_results(self, index):
        results = index.search("foo~1")
        assert len(results) == 0

    def test_edit_distance_two_results(self, index):
        results = index.search("plont~1")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"plant"}
        assert set(results[1]["match_data"].metadata.keys()) == {"plant"}


class TestSearchByField:
    def test_search_by_field_unknown_field(self, index):
        with pytest.raises(QueryParseError):
            index.search("unknown-field:plant")

    def test_search_by_field_no_results(self, index):
        results = index.search("title:candlestick")
        assert len(results) == 0

    def test_search_by_field_results(self, index):
        results = index.search("title:plant")

        assert len(results) == 1
        assert results[0]["ref"] == "b"
        assert set(results[0]["match_data"].metadata.keys()) == {"plant"}


class TestTermBoosts:
    def test_term_boosts_no_results(self, index):
        results = index.search("foo^10")
        assert len(results) == 0

    def test_term_boosts_results(self, index):
        results = index.search("scarlett candlestick^5")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"a", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"candlestick"}
        assert set(results[1]["match_data"].metadata.keys()) == {"scarlett"}


class TestTypeaheadStyleSearch:
    def test_typeahead_no_results(self, index):
        query = Query(index.fields)
        query.clause("xyz", boost=100, use_pipeline=True)
        query.clause(
            "xyz", boost=10, use_pipeline=False, wildcard=Query.WILDCARD_TRAILING
        )
        query.clause("xyz", boost=1, edit_distance=1)

        results = index.query(query)
        assert len(results) == 0

    def test_typeahead_results(self, index):
        query = Query(index.fields)
        query.clause("pl", boost=100, use_pipeline=True)
        query.clause(
            "pl", boost=10, use_pipeline=False, wildcard=Query.WILDCARD_TRAILING
        )
        query.clause("pl", boost=1, edit_distance=1)

        results = index.query(query)

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"plumb", "plant"}
        assert set(results[1]["match_data"].metadata.keys()) == {"plumb", "plant"}


class TestTermPresence:
    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_prohibited_match_excludes_prohibited_result(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("candlestick", presence=QueryPresence.PROHIBITED)
            query.term("green", presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search("-candlestick green")

        assert len(results) == 2
        assert {r["ref"] for r in results} == {"b", "c"}
        assert set(results[0]["match_data"].metadata.keys()) == {"green"}
        assert set(results[1]["match_data"].metadata.keys()) == {"green"}

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_only_prohibited_match_yields_no_results(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("green", presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search("-green")

        assert len(results) == 0

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_negated_query_no_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("qwertyuiop", presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search("-qwertyuiop")

        assert len(results) == 3

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_negated_query_some_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search("-plant")

        assert len(results) == 1
        assert results[0]["score"] == 0

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_field_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.PROHIBITED, fields=["title"])
            query.term("plumb", presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search("-title:plant plumb")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"plumb"}

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_required_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("candlestick", presence=QueryPresence.REQUIRED)
            query.term("green", presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search("+candlestick green")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"candlestick", "green"}

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_two_required_matches_yields_no_results(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("mustard", presence=QueryPresence.REQUIRED)
            query.term("plant", presence=QueryPresence.REQUIRED)
            results = index.query(query)
        else:
            results = index.search("+mustard +plant")

        assert len(results) == 0

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_required_term_not_matching_yields_no_results(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("qwertyuiop", presence=QueryPresence.REQUIRED)
            query.term("green", presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search("+qwertyuiop green")

        assert len(results) == 0

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_required_term_on_field_matches(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.REQUIRED, fields=["title"])
            query.term("green", presence=QueryPresence.OPTIONAL)
            results = index.query(query)
        else:
            results = index.search("+title:plant green")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"plant", "green"}
        assert results[0]["ref"] == "b"

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_required_terms_on_field_and_non_field_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.REQUIRED, fields=["title"])
            query.term("green", presence=QueryPresence.REQUIRED)
            results = index.query(query)
        else:
            results = index.search("+title:plant +green")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"plant", "green"}
        assert results[0]["ref"] == "b"

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_required_terms_on_different_fields_match(self, index, query_or_search):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.REQUIRED, fields=["title"])
            query.term("study", presence=QueryPresence.REQUIRED, fields=["body"])
            results = index.query(query)
        else:
            results = index.search("+title:plant +body:study")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"studi", "plant"}
        assert results[0]["ref"] == "b"

    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_combined_required_optional_and_prohibited_match(
        self, index, query_or_search
    ):
        if query_or_search == "query":
            query = Query(index.fields)
            query.term("plant", presence=QueryPresence.REQUIRED)
            query.term("green", presence=QueryPresence.OPTIONAL)
            query.term("office", presence=QueryPresence.PROHIBITED)
            results = index.query(query)
        else:
            results = index.search("+plant green -office")

        assert len(results) == 1
        assert set(results[0]["match_data"].metadata.keys()) == {"green", "plant"}
        assert results[0]["ref"] == "b"


class TestBuildTimeFieldBoost:
    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_no_query_boosts_build_boost_ranks_higher(self, documents, query_or_search):
        idx = lunr(
            ref="id",
            fields=["title", {"field_name": "body", "boost": 10}],
            documents=documents,
        )

        if query_or_search == "search":
            results = idx.search("professor")
        else:
            query = Query(idx.fields)
            query.term("professor")
            results = idx.query(query)

        assert results[0]["ref"] == "b"


class TestBuildTimeDocumentBoost:
    @pytest.mark.parametrize("query_or_search", ["query", "search"])
    def test_no_query_boosts_document_boost_ranks_higher(
        self, documents, query_or_search
    ):
        documents_with_boost = [
            (document, {"boost": 10 if document["id"] == "c" else 1})
            for document in documents
        ]
        idx = lunr(ref="id", fields=("title", "body"), documents=documents_with_boost)

        if query_or_search == "search":
            results = idx.search("plumb")
        else:
            query = Query(idx.fields)
            query.term("plumb")
            results = idx.query(query)

        assert results[0]["ref"] == "c"



print("import: 'lunr'")
import lunr

print("import: 'lunr.builder'")
import lunr.builder

print("import: 'lunr.exceptions'")
import lunr.exceptions

print("import: 'lunr.field_ref'")
import lunr.field_ref

print("import: 'lunr.idf'")
import lunr.idf

print("import: 'lunr.index'")
import lunr.index

print("import: 'lunr.languages'")
import lunr.languages

print("import: 'lunr.languages.stemmer'")
import lunr.languages.stemmer

print("import: 'lunr.languages.trimmer'")
import lunr.languages.trimmer

print("import: 'lunr.match_data'")
import lunr.match_data

print("import: 'lunr.pipeline'")
import lunr.pipeline

print("import: 'lunr.query'")
import lunr.query

print("import: 'lunr.query_lexer'")
import lunr.query_lexer

print("import: 'lunr.query_parser'")
import lunr.query_parser

print("import: 'lunr.stemmer'")
import lunr.stemmer

print("import: 'lunr.stop_word_filter'")
import lunr.stop_word_filter

print("import: 'lunr.token'")
import lunr.token

print("import: 'lunr.token_set'")
import lunr.token_set

print("import: 'lunr.token_set_builder'")
import lunr.token_set_builder

print("import: 'lunr.tokenizer'")
import lunr.tokenizer

print("import: 'lunr.trimmer'")
import lunr.trimmer

print("import: 'lunr.utils'")
import lunr.utils

print("import: 'lunr.vector'")
import lunr.vector




from lunr.utils import CompleteSet


class TestCompleteSet:
    def test_always_contains_other_element(self):
        assert "foo" in CompleteSet()

    def test_intersection_returns_other(self):
        cs = CompleteSet({"bar"})
        assert cs.intersection({"foo"}) == {"foo"}

    def test_union_returns_other(self):
        cs = CompleteSet({"bar"})
        assert cs.union({"foo"}) == {"foo"}



from __future__ import unicode_literals

"""
Implementation of Porter Stemming Algorithm from
https://tartarus.org/martin/PorterStemmer/python.txt

Note: The Python implementation returns different results than the JS
version:
- Stemming "lay" returns "lai" in Python, but "lay" in JS
- Stemming "try" returns "try" in Python, but "tri" in JS


Porter Stemming Algorithm
This is the Porter stemming algorithm, ported to Python from the
version coded up in ANSI C by the author. It may be be regarded
as canonical, in that it follows the algorithm presented in

Porter, 1980, An algorithm for suffix stripping, Program, Vol. 14,
no. 3, pp 130-137,

only differing from it at the points maked --DEPARTURE-- below.

See also http://www.tartarus.org/~martin/PorterStemmer

The algorithm as described in the paper could be exactly replicated
by adjusting the points of DEPARTURE, but this is barely necessary,
because (a) the points of DEPARTURE are definitely improvements, and
(b) no encoding of the Porter stemmer I have seen is anything like
as exact as this version, even with the points of DEPARTURE!

Vivake Gupta (v@nano.com)

Release 1: January 2001

Further adjustments by Santiago Bruno (bananabruno@gmail.com)
to allow word input not restricted to one word per line, leading
to:

release 2: July 2008
"""

from lunr.pipeline import Pipeline


class PorterStemmer:
    def __init__(self):
        """The main part of the stemming algorithm starts here.
        b is a buffer holding a word to be stemmed. The letters are in b[k0],
        b[k0+1] ... ending at b[k]. In fact k0 = 0 in this demo program. k is
        readjusted downwards as the stemming progresses. Zero termination is
        not in fact used in the algorithm.

        Note that only lower case sequences are stemmed. Forcing to lower case
        should be done before stem(...) is called.
        """

        self.b = ""  # buffer for word to be stemmed
        self.k = 0
        self.k0 = 0
        self.j = 0  # j is a general offset into the string

    def cons(self, i):
        """cons(i) is TRUE <=> b[i] is a consonant."""
        if (
            self.b[i] == "a"
            or self.b[i] == "e"
            or self.b[i] == "i"
            or self.b[i] == "o"
            or self.b[i] == "u"
        ):
            return 0
        if self.b[i] == "y":
            if i == self.k0:
                return 1
            else:
                return not self.cons(i - 1)
        return 1

    def m(self):
        """m() measures the number of consonant sequences between k0 and j.
        if c is a consonant sequence and v a vowel sequence, and <..>
        indicates arbitrary presence,

           <c><v>       gives 0
           <c>vc<v>     gives 1
           <c>vcvc<v>   gives 2
           <c>vcvcvc<v> gives 3
           ....
        """
        n = 0
        i = self.k0
        while 1:
            if i > self.j:
                return n
            if not self.cons(i):
                break
            i = i + 1
        i = i + 1
        while 1:
            while 1:
                if i > self.j:
                    return n
                if self.cons(i):
                    break
                i = i + 1
            i = i + 1
            n = n + 1
            while 1:
                if i > self.j:
                    return n
                if not self.cons(i):
                    break
                i = i + 1
            i = i + 1

    def vowelinstem(self):
        """vowelinstem() is TRUE <=> k0,...j contains a vowel"""
        for i in range(self.k0, self.j + 1):
            if not self.cons(i):
                return 1
        return 0

    def doublec(self, j):
        """doublec(j) is TRUE <=> j,(j-1) contain a double consonant."""
        if j < (self.k0 + 1):
            return 0
        if self.b[j] != self.b[j - 1]:
            return 0
        return self.cons(j)

    def cvc(self, i):
        """cvc(i) is TRUE <=> i-2,i-1,i has the form consonant - vowel - consonant
        and also if the second c is not w,x or y. this is used when trying to
        restore an e at the end of a short  e.g.

           cav(e), lov(e), hop(e), crim(e), but
           snow, box, tray.
        """
        if (
            i < (self.k0 + 2)
            or not self.cons(i)
            or self.cons(i - 1)
            or not self.cons(i - 2)
        ):
            return 0
        ch = self.b[i]
        if ch == "w" or ch == "x" or ch == "y":
            return 0
        return 1

    def ends(self, s):
        """ends(s) is TRUE <=> k0,...k ends with the string s."""
        length = len(s)
        if s[length - 1] != self.b[self.k]:  # tiny speed-up
            return 0
        if length > (self.k - self.k0 + 1):
            return 0
        if self.b[self.k - length + 1 : self.k + 1] != s:
            return 0
        self.j = self.k - length
        return 1

    def setto(self, s):
        """setto(s) sets (j+1),...k to the characters in the string s, readjusting k."""
        length = len(s)
        self.b = self.b[: self.j + 1] + s + self.b[self.j + length + 1 :]
        self.k = self.j + length

    def r(self, s):
        """r(s) is used further down."""
        if self.m() > 0:
            self.setto(s)

    def step1ab(self):
        """step1ab() gets rid of plurals and -ed or -ing. e.g.

           caresses  ->  caress
           ponies    ->  poni
           ties      ->  ti
           caress    ->  caress
           cats      ->  cat

           feed      ->  feed
           agreed    ->  agree
           disabled  ->  disable

           matting   ->  mat
           mating    ->  mate
           meeting   ->  meet
           milling   ->  mill
           messing   ->  mess

           meetings  ->  meet
        """
        if self.b[self.k] == "s":
            if self.ends("sses"):
                self.k = self.k - 2
            elif self.ends("ies"):
                self.setto("i")
            elif self.b[self.k - 1] != "s":
                self.k = self.k - 1
        if self.ends("eed"):
            if self.m() > 0:
                self.k = self.k - 1
        elif (self.ends("ed") or self.ends("ing")) and self.vowelinstem():
            self.k = self.j
            if self.ends("at"):
                self.setto("ate")
            elif self.ends("bl"):
                self.setto("ble")
            elif self.ends("iz"):
                self.setto("ize")
            elif self.doublec(self.k):
                self.k = self.k - 1
                ch = self.b[self.k]
                if ch == "l" or ch == "s" or ch == "z":
                    self.k = self.k + 1
            elif self.m() == 1 and self.cvc(self.k):
                self.setto("e")

    def step1c(self):
        """step1c() turns terminal y to i when there is another vowel in the stem."""
        if self.ends("y") and self.vowelinstem():
            self.b = self.b[: self.k] + "i" + self.b[self.k + 1 :]

    def step2(self):
        """step2() maps double suffices to single ones.
        so -ization ( = -ize plus -ation) maps to -ize etc. note that the
        string before the suffix must give m() > 0.
        """
        if self.b[self.k - 1] == "a":
            if self.ends("ational"):
                self.r("ate")
            elif self.ends("tional"):
                self.r("tion")
        elif self.b[self.k - 1] == "c":
            if self.ends("enci"):
                self.r("ence")
            elif self.ends("anci"):
                self.r("ance")
        elif self.b[self.k - 1] == "e":
            if self.ends("izer"):
                self.r("ize")
        elif self.b[self.k - 1] == "l":
            if self.ends("bli"):
                self.r("ble")  # --DEPARTURE--
            # To match the published algorithm, replace this phrase with
            #   if self.ends("abli"):      self.r("able")
            elif self.ends("alli"):
                self.r("al")
            elif self.ends("entli"):
                self.r("ent")
            elif self.ends("eli"):
                self.r("e")
            elif self.ends("ousli"):
                self.r("ous")
        elif self.b[self.k - 1] == "o":
            if self.ends("ization"):
                self.r("ize")
            elif self.ends("ation"):
                self.r("ate")
            elif self.ends("ator"):
                self.r("ate")
        elif self.b[self.k - 1] == "s":
            if self.ends("alism"):
                self.r("al")
            elif self.ends("iveness"):
                self.r("ive")
            elif self.ends("fulness"):
                self.r("ful")
            elif self.ends("ousness"):
                self.r("ous")
        elif self.b[self.k - 1] == "t":
            if self.ends("aliti"):
                self.r("al")
            elif self.ends("iviti"):
                self.r("ive")
            elif self.ends("biliti"):
                self.r("ble")
        elif self.b[self.k - 1] == "g":  # --DEPARTURE--
            if self.ends("logi"):
                self.r("log")
        # To match the published algorithm, delete this phrase

    def step3(self):
        """step3() dels with -ic-, -full, -ness etc. similar strategy to step2."""
        if self.b[self.k] == "e":
            if self.ends("icate"):
                self.r("ic")
            elif self.ends("ative"):
                self.r("")
            elif self.ends("alize"):
                self.r("al")
        elif self.b[self.k] == "i":
            if self.ends("iciti"):
                self.r("ic")
        elif self.b[self.k] == "l":
            if self.ends("ical"):
                self.r("ic")
            elif self.ends("ful"):
                self.r("")
        elif self.b[self.k] == "s":
            if self.ends("ness"):
                self.r("")

    def step4(self):
        """step4() takes off -ant, -ence etc., in context <c>vcvc<v>."""
        if self.b[self.k - 1] == "a":
            if self.ends("al"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "c":
            if self.ends("ance"):
                pass
            elif self.ends("ence"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "e":
            if self.ends("er"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "i":
            if self.ends("ic"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "l":
            if self.ends("able"):
                pass
            elif self.ends("ible"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "n":
            if self.ends("ant"):
                pass
            elif self.ends("ement"):
                pass
            elif self.ends("ment"):
                pass
            elif self.ends("ent"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "o":
            if self.ends("ion") and (self.b[self.j] == "s" or self.b[self.j] == "t"):
                pass
            elif self.ends("ou"):
                pass
            # takes care of -ous
            else:
                return
        elif self.b[self.k - 1] == "s":
            if self.ends("ism"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "t":
            if self.ends("ate"):
                pass
            elif self.ends("iti"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "u":
            if self.ends("ous"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "v":
            if self.ends("ive"):
                pass
            else:
                return
        elif self.b[self.k - 1] == "z":
            if self.ends("ize"):
                pass
            else:
                return
        else:
            return
        if self.m() > 1:
            self.k = self.j

    def step5(self):
        """step5() removes a final -e if m() > 1, and changes -ll to -l if
        m() > 1.
        """
        self.j = self.k
        if self.b[self.k] == "e":
            a = self.m()
            if a > 1 or (a == 1 and not self.cvc(self.k - 1)):
                self.k = self.k - 1
        if self.b[self.k] == "l" and self.doublec(self.k) and self.m() > 1:
            self.k = self.k - 1

    def stem(self, p, metadata=None):
        """In stem(p,i,j), p is a char pointer, and the string to be stemmed
        is from p[i] to p[j] inclusive. Typically i is zero and j is the
        offset to the last character of a string, (p[j+1] == '\0'). The
        stemmer adjusts the characters p[i] ... p[j] and returns the new
        end-point of the string, k. Stemming never increases word length, so
        i <= k <= j. To turn the stemmer into a module, declare 'stem' as
        extern, and delete the remainder of this file.
        """
        # TODO: removed i and j from the original implementation
        # to comply with the `token.update` API
        i = 0
        j = len(p) - 1
        # copy the parameters into statics
        self.b = p
        self.k = j
        self.k0 = i
        if self.k <= self.k0 + 1:
            return self.b  # --DEPARTURE--

        # With this line, strings of length 1 or 2 don't go through the
        # stemming process, although no mention is made of this in the
        # published algorithm. Remove the line to match the published
        # algorithm.

        self.step1ab()
        self.step1c()
        self.step2()
        self.step3()
        self.step4()
        self.step5()
        return self.b[self.k0 : self.k + 1]


porter_stemmer = PorterStemmer()


def stemmer(token, i=None, tokens=None):
    """Wrapper around the PorterStemmer for inclusion in pipeline.

    Args:
        language (str): ISO-639-1 code of the language.
        token (lunr.Token): The token to stem.
        i (int): The index of the token in a set.
        tokens (list): A list of tokens representing the set.
    """
    return token.update(porter_stemmer.stem)


Pipeline.register_function(stemmer, "stemmer")



from lunr.token import Token


def test_str_repr():
    token = Token("foo")
    assert str(token) == "foo"
    assert repr(token) == '<Token "foo">'


class TestMetadata:
    def test_can_attach_arbitrary_metadata(self):
        token = Token("foo", {"length": 3})
        assert token.metadata["length"] == 3

    def test_can_update_token_value(self):
        token = Token("foo", {"length": 3})
        token.update(lambda s, m: s.upper())

        assert str(token) == "FOO"

    def test_metadata_is_yielded_when_updating(self):
        # TODO: unsure what this test is asserting, a language feature?
        pass


class TestClone:
    def setup_method(self, method):
        self.token = Token("foo", {"bar": True})

    def test_clones_value(self):
        assert str(self.token) == str(self.token.clone())

    def test_clones_metadata(self):
        assert self.token.metadata == self.token.clone().metadata

    def test_clone_and_modify(self):
        clone = self.token.clone(lambda s, m: s.upper())

        assert str(clone) == "FOO"
        self.token.metadata == clone.metadata



from lunr.query_lexer import QueryLexer


def _lex(string):
    lexer = QueryLexer(string)
    lexer.run()
    return lexer


class TestQueryLexer:
    def test_single_term_produces_one_lexeme(self):
        lexer = _lex("foo")
        assert len(lexer.lexemes) == 1
        lexeme = lexer.lexemes[0]
        assert lexeme["type"] == QueryLexer.TERM
        assert lexeme["string"] == "foo"
        assert lexeme["start"] == 0
        assert lexeme["end"] == 3

    def test_term_escape_character(self):
        lexer = _lex("foo\\:bar")
        assert len(lexer.lexemes) == 1
        lexeme = lexer.lexemes[0]
        assert lexeme["type"] == QueryLexer.TERM
        assert lexeme["string"] == "foo:bar"
        assert lexeme["start"] == 0
        assert lexeme["end"] == 8

    def test_multiple_terms(self):
        lexer = _lex("foo bar")
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes
        assert foo_lexeme["type"] == bar_lexeme["type"] == QueryLexer.TERM
        assert foo_lexeme["string"] == "foo"
        assert bar_lexeme["string"] == "bar"
        assert foo_lexeme["start"] == 0
        assert bar_lexeme["start"] == 4
        assert foo_lexeme["end"] == 3
        assert bar_lexeme["end"] == 7

    def test_separator_length_greater_than_one(self):
        lexer = _lex("foo    bar")
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes
        assert foo_lexeme["type"] == bar_lexeme["type"] == QueryLexer.TERM
        assert foo_lexeme["string"] == "foo"
        assert bar_lexeme["string"] == "bar"
        assert foo_lexeme["start"] == 0
        assert bar_lexeme["start"] == 7
        assert foo_lexeme["end"] == 3
        assert bar_lexeme["end"] == 10

    def test_hyphen_is_considered_a_separator(self):
        lexer = _lex("foo-bar")
        assert len(lexer.lexemes) == 2

    def test_term_with_field(self):
        lexer = _lex("title:foo")
        assert len(lexer.lexemes) == 2
        field_lexeme, term_lexeme = lexer.lexemes
        assert field_lexeme["type"] == QueryLexer.FIELD
        assert term_lexeme["type"] == QueryLexer.TERM
        assert field_lexeme["string"] == "title"
        assert term_lexeme["string"] == "foo"
        assert field_lexeme["start"] == 0
        assert term_lexeme["start"] == 6
        assert field_lexeme["end"] == 5
        assert term_lexeme["end"] == 9

    def test_term_with_field_with_escape_character(self):
        lexer = _lex("ti\\:tle:foo")
        assert len(lexer.lexemes) == 2
        field_lexeme, term_lexeme = lexer.lexemes
        assert field_lexeme["type"] == QueryLexer.FIELD
        assert term_lexeme["type"] == QueryLexer.TERM
        assert field_lexeme["string"] == "ti:tle"
        assert term_lexeme["string"] == "foo"
        assert field_lexeme["start"] == 0
        assert term_lexeme["start"] == 8
        assert field_lexeme["end"] == 7
        assert term_lexeme["end"] == 11

    def test_term_with_edit_distance(self):
        lexer = _lex("foo~2")
        assert len(lexer.lexemes) == 2
        term_lexeme, edit_distance_lexeme = lexer.lexemes
        assert term_lexeme["type"] == QueryLexer.TERM
        assert edit_distance_lexeme["type"] == QueryLexer.EDIT_DISTANCE
        assert term_lexeme["string"] == "foo"
        assert edit_distance_lexeme["string"] == "2"
        assert term_lexeme["start"] == 0
        assert edit_distance_lexeme["start"] == 4
        assert term_lexeme["end"] == 3
        assert edit_distance_lexeme["end"] == 5

    def test_term_with_boost(self):
        lexer = _lex("foo^10")
        assert len(lexer.lexemes) == 2
        term_lexeme, boost_lexeme = lexer.lexemes
        assert term_lexeme["type"] == QueryLexer.TERM
        assert boost_lexeme["type"] == QueryLexer.BOOST
        assert term_lexeme["string"] == "foo"
        assert boost_lexeme["string"] == "10"
        assert term_lexeme["start"] == 0
        assert boost_lexeme["start"] == 4
        assert term_lexeme["end"] == 3
        assert boost_lexeme["end"] == 6

    def test_term_with_field_boost_and_edit_distance(self):
        lexer = _lex("title:foo^10~5")
        assert len(lexer.lexemes) == 4
        field_lexeme, term_lexeme, boost_lexeme, edit_distance_lexeme = lexer.lexemes
        assert field_lexeme["type"] == QueryLexer.FIELD
        assert term_lexeme["type"] == QueryLexer.TERM
        assert boost_lexeme["type"] == QueryLexer.BOOST
        assert edit_distance_lexeme["type"] == QueryLexer.EDIT_DISTANCE

        assert field_lexeme["string"] == "title"
        assert term_lexeme["string"] == "foo"
        assert boost_lexeme["string"] == "10"
        assert edit_distance_lexeme["string"] == "5"

        assert field_lexeme["start"] == 0
        assert term_lexeme["start"] == 6
        assert boost_lexeme["start"] == 10
        assert edit_distance_lexeme["start"] == 13

        assert field_lexeme["end"] == 5
        assert term_lexeme["end"] == 9
        assert boost_lexeme["end"] == 12
        assert edit_distance_lexeme["end"] == 14

    def test_single_term_with_hyphen_produces_two_lexemes(self):
        """Embedded hyphens should not be confused with presence operators."""
        lexer = _lex("foo-bar")
        assert len(lexer.lexemes) == 2
        foo_lexeme, bar_lexeme = lexer.lexemes

        assert foo_lexeme["type"] == QueryLexer.TERM
        assert foo_lexeme["string"] == "foo"
        assert foo_lexeme["start"] == 0
        assert foo_lexeme["end"] == 3

        assert bar_lexeme["type"] == QueryLexer.TERM
        assert bar_lexeme["string"] == "bar"
        assert bar_lexeme["start"] == 4
        assert bar_lexeme["end"] == 7

    def test_single_term_with_presence_produces_two_lexemes(self):
        lexer = _lex("+foo")
        assert len(lexer.lexemes) == 2
        presence_lexeme, term_lexeme = lexer.lexemes

        assert presence_lexeme["type"] == QueryLexer.PRESENCE
        assert presence_lexeme["string"] == "+"
        assert presence_lexeme["start"] == 0
        assert presence_lexeme["end"] == 1

        assert term_lexeme["type"] == QueryLexer.TERM
        assert term_lexeme["string"] == "foo"
        assert term_lexeme["start"] == 1
        assert term_lexeme["end"] == 4

    def test_multiple_terms_with_presence_produces_four_lexemes(self):
        lexer = _lex("+foo +bar")
        assert len(lexer.lexemes) == 4
        (
            foo_presence_lexeme,
            foo_term_lexeme,
            bar_presence_lexeme,
            bar_term_lexeme,
        ) = lexer.lexemes

        assert foo_term_lexeme["type"] == QueryLexer.TERM
        assert foo_term_lexeme["string"] == "foo"
        assert foo_term_lexeme["start"] == 1
        assert foo_term_lexeme["end"] == 4

        assert foo_presence_lexeme["type"] == QueryLexer.PRESENCE
        assert foo_presence_lexeme["string"] == "+"
        assert foo_presence_lexeme["start"] == 0
        assert foo_presence_lexeme["end"] == 1

        assert bar_term_lexeme["type"] == QueryLexer.TERM
        assert bar_term_lexeme["string"] == "bar"
        assert bar_term_lexeme["start"] == 6
        assert bar_term_lexeme["end"] == 9

        assert bar_presence_lexeme["type"] == QueryLexer.PRESENCE
        assert bar_presence_lexeme["string"] == "+"
        assert bar_presence_lexeme["start"] == 5
        assert bar_presence_lexeme["end"] == 6

    def test_multiple_terms_with_presence_and_fuzz(self):
        lexer = _lex("+foo~1 +bar")
        assert len(lexer.lexemes) == 5

        (
            foo_presence_lexeme,
            foo_term_lexeme,
            foo_fuzz_lexeme,
            bar_presence_lexeme,
            bar_term_lexeme,
        ) = lexer.lexemes

        assert foo_presence_lexeme["type"] == QueryLexer.PRESENCE
        assert foo_presence_lexeme["string"] == "+"
        assert foo_presence_lexeme["start"] == 0
        assert foo_presence_lexeme["end"] == 1

        assert foo_term_lexeme["type"] == QueryLexer.TERM
        assert foo_term_lexeme["string"] == "foo"
        assert foo_term_lexeme["start"] == 1
        assert foo_term_lexeme["end"] == 4

        assert foo_fuzz_lexeme["type"] == QueryLexer.EDIT_DISTANCE
        assert foo_fuzz_lexeme["string"] == "1"
        assert foo_fuzz_lexeme["start"] == 5
        assert foo_fuzz_lexeme["end"] == 6

        assert bar_presence_lexeme["type"] == QueryLexer.PRESENCE
        assert bar_presence_lexeme["string"] == "+"
        assert bar_presence_lexeme["start"] == 7
        assert bar_presence_lexeme["end"] == 8

        assert bar_term_lexeme["type"] == QueryLexer.TERM
        assert bar_term_lexeme["string"] == "bar"
        assert bar_term_lexeme["start"] == 8
        assert bar_term_lexeme["end"] == 11



from __future__ import unicode_literals, division

from collections import defaultdict

from builtins import str, dict  # noqa

from lunr.pipeline import Pipeline
from lunr.tokenizer import Tokenizer
from lunr.token_set import TokenSet
from lunr.field_ref import FieldRef
from lunr.index import Index
from lunr.vector import Vector
from lunr.idf import idf as Idf


class Field:
    """Represents a field with boost and extractor functions."""

    def __init__(self, field_name, boost=1, extractor=None):
        self.name = field_name
        self.boost = boost
        self.extractor = extractor

    def __repr__(self):
        return '<Field "{0.name}" boost="{0.boost}">'.format(self)

    def __hash__(self):
        return hash(self.name)


class Builder:
    """Performs indexing on a set of documents and returns instances of
    lunr.Index ready for querying.

    All configuration of the index is done via the builder, the fields to
    index, the document reference, the text processing pipeline and document
    scoring parameters are all set on the builder before indexing.
    """

    def __init__(self):
        self._ref = "id"
        self._fields = {}
        self.inverted_index = {}
        self.field_term_frequencies = {}
        self.field_lengths = {}
        self.pipeline = Pipeline()
        self.search_pipeline = Pipeline()
        self._documents = {}
        self.document_count = 0
        self._b = 0.75
        self._k1 = 1.2
        self.term_index = 0
        self.metadata_whitelist = []

    def ref(self, ref):
        """Sets the document field used as the document reference.

        Every document must have this field. The type of this field in the
        document should be a string, if it is not a string it will be coerced
        into a string by calling `str`.

        The default ref is 'id'. The ref should _not_ be changed during
        indexing, it should be set before any documents are added to the index.
        Changing it during indexing can lead to inconsistent results.

        """
        self._ref = ref

    def field(self, field_name, boost=1, extractor=None):
        """Adds a field to the list of document fields that will be indexed.

        Every document being indexed should have this field. None values for
        this field in indexed documents will not cause errors but will limit
        the chance of that document being retrieved by searches.

        All fields should be added before adding documents to the index. Adding
        fields after a document has been indexed will have no effect on already
        indexed documents.

        Fields can be boosted at build time. This allows terms within that
        field to have more importance on search results. Use a field boost to
        specify that matches within one field are more important that other
        fields.

        Args:
            field_name (str): Name of the field to be added, must not include
                a forward slash '/'.
            boost (int): Optional boost factor to apply to field.
            extractor (callable): Optional function to extract a field from
                the document.

        Raises:
            ValueError: If the field name contains a `/`.
        """
        if "/" in field_name:
            raise ValueError("Field {} contains illegal character `/`")

        self._fields[field_name] = Field(field_name, boost, extractor)

    def b(self, number):
        """A parameter to tune the amount of field length normalisation that is
        applied when calculating relevance scores.

        A value of 0 will completely disable any normalisation and a value of 1
        will fully normalise field lengths. The default is 0.75. Values of b
        will be clamped to the range 0 - 1.
        """
        if number < 0:
            self._b = 0
        elif number > 1:
            self._b = 1
        else:
            self._b = number

    def k1(self, number):
        """ A parameter that controls the speed at which a rise in term
        frequency results in term frequency saturation.

        The default value is 1.2. Setting this to a higher value will give
        slower saturation levels, a lower value will result in quicker
        saturation.
        """
        self._k1 = number

    def add(self, doc, attributes=None):
        """Adds a document to the index.

        Before adding documents to the index it should have been fully
        setup, with the document ref and all fields to index already having
        been specified.

        The document must have a field name as specified by the ref (by default
        this is 'id') and it should have all fields defined for indexing,
        though None values will not cause errors.

        Args:
            - doc (dict): The document to be added to the index.
            - attributes (dict, optional): A set of attributes corresponding
            to the document, currently a single `boost` -> int will be
            taken into account.
        """
        doc_ref = str(doc[self._ref])
        self._documents[doc_ref] = attributes or {}
        self.document_count += 1

        for field_name, field in self._fields.items():
            extractor = field.extractor
            field_value = doc[field_name] if extractor is None else extractor(doc)
            tokens = Tokenizer(field_value)
            terms = self.pipeline.run(tokens)
            field_ref = FieldRef(doc_ref, field_name)
            field_terms = defaultdict(int)

            # TODO: field_refs are casted to strings in JS, should we allow
            # FieldRef as keys?
            self.field_term_frequencies[str(field_ref)] = field_terms
            self.field_lengths[str(field_ref)] = len(terms)

            for term in terms:
                # TODO: term is a Token, should we allow Tokens as keys?
                term_key = str(term)

                field_terms[term_key] += 1
                if term_key not in self.inverted_index:
                    posting = {_field_name: {} for _field_name in self._fields}
                    posting["_index"] = self.term_index
                    self.term_index += 1
                    self.inverted_index[term_key] = posting

                if doc_ref not in self.inverted_index[term_key][field_name]:
                    self.inverted_index[term_key][field_name][doc_ref] = defaultdict(
                        list
                    )

                for metadata_key in self.metadata_whitelist:
                    metadata = term.metadata[metadata_key]
                    self.inverted_index[term_key][field_name][doc_ref][
                        metadata_key
                    ].append(metadata)

    def build(self):
        """Builds the index, creating an instance of `lunr.Index`.

        This completes the indexing process and should only be called once all
        documents have been added to the index.
        """
        self._calculate_average_field_lengths()
        self._create_field_vectors()
        self._create_token_set()

        return Index(
            inverted_index=self.inverted_index,
            field_vectors=self.field_vectors,
            token_set=self.token_set,
            fields=list(self._fields.keys()),
            pipeline=self.search_pipeline,
        )

    def _create_token_set(self):
        """Creates a token set of all tokens in the index using `lunr.TokenSet`
        """
        self.token_set = TokenSet.from_list(sorted(list(self.inverted_index.keys())))

    def _calculate_average_field_lengths(self):
        """Calculates the average document length for this index"""
        accumulator = defaultdict(int)
        documents_with_field = defaultdict(int)

        for field_ref, length in self.field_lengths.items():
            _field_ref = FieldRef.from_string(field_ref)
            field = _field_ref.field_name

            documents_with_field[field] += 1
            accumulator[field] += length

        for field_name in self._fields:
            accumulator[field_name] /= documents_with_field[field_name]

        self.average_field_length = accumulator

    def _create_field_vectors(self):
        """Builds a vector space model of every document using lunr.Vector."""
        field_vectors = {}
        term_idf_cache = {}

        for field_ref, term_frequencies in self.field_term_frequencies.items():
            _field_ref = FieldRef.from_string(field_ref)
            field_name = _field_ref.field_name
            field_length = self.field_lengths[field_ref]
            field_vector = Vector()
            field_boost = self._fields[field_name].boost
            doc_boost = self._documents[_field_ref.doc_ref].get("boost", 1)

            for term, tf in term_frequencies.items():
                term_index = self.inverted_index[term]["_index"]

                if term not in term_idf_cache:
                    idf = Idf(self.inverted_index[term], self.document_count)
                    term_idf_cache[term] = idf
                else:
                    idf = term_idf_cache[term]

                score = (
                    idf
                    * ((self._k1 + 1) * tf)
                    / (
                        self._k1
                        * (
                            1
                            - self._b
                            + self._b
                            * (field_length / self.average_field_length[field_name])
                        )
                        + tf
                    )
                )
                score *= field_boost
                score *= doc_boost
                score_with_precision = round(score, 3)

                field_vector.insert(term_index, score_with_precision)

            field_vectors[field_ref] = field_vector

        self.field_vectors = field_vectors

    def use(self, fn, *args, **kwargs):
        """Applies a plugin to the index builder.

        A plugin is a function that is called with the index builder as its
        context. Plugins can be used to customise or extend the behaviour of
        the index in some way.

        A plugin is just a function, that encapsulated the custom behaviour
        that should be applied when building the index. The plugin function
        will be called with the index builder as its argument, additional
        arguments can also be passed when calling use.
        """
        fn(self, *args, **kwargs)



from __future__ import unicode_literals

from builtins import str

import pytest

from lunr.builder import Builder
from lunr.token_set import TokenSet
from lunr.index import Index
from lunr.vector import Vector


def _assert_deep_keys(dict_, keys):
    d = dict_
    for key in keys.split("."):
        d_keys_as_str = [str(k) for k in d]
        assert key in d_keys_as_str
        d = d[key]


class TestBuilderBuild:
    def setup_method(self, method):
        self.builder = Builder()
        doc = {"id": "id", "title": "test", "body": "missing"}

        self.builder.ref("id")
        self.builder.field("title")
        self.builder.add(doc)
        self.index = self.builder.build()

    def test_adds_tokens_to_inverted_index(self):
        _assert_deep_keys(self.builder.inverted_index, "test.title.id")

    def test_builds_vector_space_of_the_document_fields(self):
        assert "title/id" in self.builder.field_vectors
        assert isinstance(self.builder.field_vectors["title/id"], Vector)

    def test_skips_fields_not_defined_for_indexing(self):
        assert "missing" not in self.builder.inverted_index

    def test_builds_a_token_set_for_the_corpus(self):
        needle = TokenSet.from_string("test")
        assert "test" in self.builder.token_set.intersect(needle).to_list()

    def test_calculates_document_count(self):
        assert self.builder.average_field_length["title"] == 1

    def test_index_is_returned(self):
        assert isinstance(self.index, Index)


class TestBuilderAdd:
    def test_builder_casts_docrefs_to_strings(self):
        self.builder = Builder()
        self.builder.ref("id")
        self.builder.field("title")

        self.builder.add(dict(id=123, title="test", body="missing"))

        _assert_deep_keys(self.builder.inverted_index, "test.title.123")

    def test_builder_metadata_whitelist_includes_metadata_in_index(self):
        self.builder = Builder()
        self.builder.ref("id")
        self.builder.field("title")
        self.builder.metadata_whitelist = ["position"]

        self.builder.add(dict(id="a", title="test", body="missing"))
        self.builder.add(dict(id="b", title="another test", body="missing"))

        assert self.builder.inverted_index["test"]["title"]["a"] == {
            "position": [[0, 4]]
        }
        assert self.builder.inverted_index["test"]["title"]["b"] == {
            "position": [[8, 4]]
        }

    def test_builder_field_raises_if_contains_slash(self):
        self.builder = Builder()

        with pytest.raises(ValueError):
            self.builder.field("foo/bar")

    def test_builder_extracts_nested_properties_from_document(self):
        self.builder = Builder()
        self.builder.field("name", extractor=lambda d: d["person"]["name"])

        self.builder.add({"id": "id", "person": {"name": "bob"}})

        assert self.builder.inverted_index["bob"]["name"]["id"] == {}

    def test_builder_field_term_frequency_and_length(self):
        self.builder = Builder()
        self.builder.ref("id")
        self.builder.field("title")

        self.builder.add(dict(id="a", title="test a testing test", body="missing"))

        assert self.builder.field_term_frequencies == {
            "title/a": {"test": 2, "a": 1, "testing": 1}
        }
        assert self.builder.field_lengths == {"title/a": 4}


class TestBuilderUse:
    def setup_method(self, method):
        self.builder = Builder()

    def test_calls_plugin_function(self):
        def plugin(*args):
            assert True

        self.builder.use(plugin)

    def test_plugin_is_called_with_builder_as_first_argument(self):
        def plugin(builder):
            assert builder is self.builder

        self.builder.use(plugin)

    def test_forwards_arguments_to_the_plugin(self):
        def plugin(builder, *args, **kwargs):
            assert args == (1, 2, 3)
            assert kwargs == {"foo": "bar"}

        self.builder.use(plugin, 1, 2, 3, foo="bar")


class TestBuilderK1:
    def test_k1_default_value(self):
        builder = Builder()
        assert builder._k1 == 1.2

    def test_k1_can_be_set(self):
        builder = Builder()
        builder.k1(1.6)
        assert builder._k1 == 1.6


class TestBuilderB:
    def test_b_default_value(self):
        builder = Builder()
        assert builder._b == 0.75

    def test_b_within_range(self):
        builder = Builder()
        builder.b(0.5)
        assert builder._b == 0.5

    def test_b_less_than_zero(self):
        builder = Builder()
        builder.b(-1)
        assert builder._b == 0

    def test_b_higher_than_one(self):
        builder = Builder()
        builder.b(1.5)
        assert builder._b == 1


class TestBuilerRef:
    def test_default_reference(self):
        builder = Builder()
        assert builder._ref == "id"

    def test_defining_a_reference_field(self):
        builder = Builder()
        builder.ref("foo")
        assert builder._ref == "foo"


class TestBuilderField:
    def test_define_fields_to_index(self):
        builder = Builder()
        builder.field("foo")
        assert len(builder._fields) == 1
        assert builder._fields["foo"].name == "foo"
        assert builder._fields["foo"].boost == 1
        assert builder._fields["foo"].extractor is None
        assert repr(builder._fields["foo"]) == '<Field "foo" boost="1">'
        assert hash(builder._fields["foo"]) == hash("foo")



from __future__ import unicode_literals

import re


def generate_trimmer(word_characters):
    """Returns a trimmer function from a string of word characters.

    TODO: lunr-languages ships with lists of word characters for each language
    I haven't found an equivalent in Python, we may need to copy it.
    """
    full_re = re.compile(r"^[^{0}]*?([{0}]+)[^{0}]*?$".format(word_characters))

    def trimmer(token, i=None, tokens=None):
        def trim(s, metadata=None):
            match = full_re.match(s)
            if match is None:
                return s
            return match.group(1)

        return token.update(trim)

    return trimmer



from __future__ import unicode_literals

import re
from builtins import str

import pytest
import six

from lunr.tokenizer import Tokenizer


class TestTokenizer:
    def test_splitting_into_tokens(self):
        tokenizer = Tokenizer("foo bar baz")
        tokens = [str(token) for token in tokenizer]

        assert tokens == ["foo", "bar", "baz"]

    def test_run_downcases_tokens(self):
        tokenizer = Tokenizer("foo BAR BAZ")
        tokens = [str(token) for token in tokenizer]

        assert tokens == ["foo", "bar", "baz"]

    def test_array_of_strings(self):
        tokenizer = Tokenizer(["foo", "bar", "baz"])
        tokens = [str(token) for token in tokenizer]

        assert tokens == ["foo", "bar", "baz"]

    def test_none_is_converted_to_empty_string(self):
        tokenizer = Tokenizer(["foo", None, "baz"])
        tokens = [str(token) for token in tokenizer]

        assert tokens == ["foo", "", "baz"]

    def test_multiple_whitespace_is_stripped(self):
        tokenizer = Tokenizer("   foo    bar   baz  ")
        tokens = [str(token) for token in tokenizer]

        assert tokens == ["foo", "bar", "baz"]

    def test_handling_null_like_arguments(self):
        assert len(Tokenizer(None)) == 0

    def test_converting_a_number_to_tokens(self):
        tokens = [str(token) for token in Tokenizer(41)]
        assert tokens == ["41"]

    def test_converting_a_boolean_to_tokens(self):
        tokens = [str(token) for token in Tokenizer(False)]
        assert tokens == ["false"]

    def test_converting_an_object_to_tokens(self):
        @six.python_2_unicode_compatible
        class Subject:
            def __str__(self):
                return "custom object"

        tokens = [str(token) for token in Tokenizer(Subject())]
        assert tokens == ["custom", "object"]

    def test_splits_strings_with_hyphens(self):
        tokens = [str(token) for token in Tokenizer("foo-bar")]
        assert tokens == ["foo", "bar"]

    def test_splits_strings_with_hyphens_and_spaces(self):
        tokens = [str(token) for token in Tokenizer("foo - bar")]
        assert tokens == ["foo", "bar"]

    def test_tracking_the_token_index(self):
        tokens = Tokenizer("foo bar")
        assert tokens[0].metadata["index"] == 0
        assert tokens[1].metadata["index"] == 1

    def test_tracking_the_token_position(self):
        tokens = Tokenizer("foo bar")
        assert tokens[0].metadata["position"] == [0, 3]
        assert tokens[1].metadata["position"] == [4, 3]

    def test_providing_additional_metadata(self):
        tokens = Tokenizer("foo bar", {"hurp": "durp"})
        assert tokens[0].metadata["hurp"] == "durp"
        assert tokens[1].metadata["hurp"] == "durp"

    @pytest.mark.parametrize("separator", [re.compile(r"[_\-]+"), lambda c: c in "_-"])
    def test_providing_separator(self, separator):
        tokens = [str(token) for token in Tokenizer("foo_bar-baz", separator=separator)]
        assert tokens == ["foo", "bar", "baz"]

    def test_tracking_token_position_with_left_hand_whitespace(self):
        tokens = Tokenizer(" foo bar")
        assert tokens[0].metadata["position"] == [1, 3]
        assert tokens[1].metadata["position"] == [5, 3]

    def test_tracking_token_position_with_right_hand_whitespace(self):
        tokens = Tokenizer("foo bar ")
        assert tokens[0].metadata["position"] == [0, 3]
        assert tokens[1].metadata["position"] == [4, 3]



from __future__ import unicode_literals

from builtins import str

import pytest

from lunr.trimmer import trimmer
from lunr.token import Token
from lunr.pipeline import Pipeline


class TestTrimmer:
    def test_latin_characters(self):
        token = Token("hello")
        assert str(trimmer(token)) == str(token)

    @pytest.mark.parametrize(
        "description, string, expected",
        [
            ("full stop", "hello.", "hello"),
            ("inner apostrophe", "it's", "it's"),
            ("trailing apostrophe", "james'", "james"),
            ("exclamation mark", "stop!", "stop"),
            ("comma", "first,", "first"),
            ("brackets", "[tag]", "tag"),
        ],
    )
    def test_punctuation(self, description, string, expected):
        token = Token(string)
        trimmed = str(trimmer(token))

        assert trimmed == expected

    def test_is_a_registered_pipeline_function(self):
        assert trimmer.label == "trimmer"
        assert Pipeline.registered_functions["trimmer"] == trimmer



import json

from lunr import lunr
from lunr.index import Index


class TestSerialization:
    def setup_method(self, method):
        documents = [
            {
                "id": "a",
                "title": "Mr. Green kills Colonel Mustard",
                "body": """Mr. Green killed Colonel Mustard in the study with the
candlestick. Mr. Green is not a very nice fellow.""",
                "word_count": 19,
            },
            {
                "id": "b",
                "title": "Plumb waters plant",
                "body": "Professor Plumb has a green plant in his study",
                "word_count": 9,
            },
            {
                "id": "c",
                "title": "Scarlett helps Professor",
                "body": """Miss Scarlett watered Professor Plumbs green plant
while he was away from his office last week.""",
                "word_count": 16,
            },
        ]

        self.idx = lunr(ref="id", fields=("title", "body"), documents=documents)

    def test_serialization(self):
        serialized_index = json.dumps(self.idx.serialize())
        loaded_index = Index.load(json.loads(serialized_index))

        assert self.idx.search("green") == loaded_index.search("green")
